/****************************************************************************
*                   mblur2.c
*
*  This module implements routines for another type of motion blur.
*
*  from Persistence of Vision(tm) Ray Tracer
*  Copyright 1996,1999 Persistence of Vision Team
*---------------------------------------------------------------------------
*  NOTICE: This source code file is provided so that users may experiment
*  with enhancements to POV-Ray and to port the software to platforms other
*  than those supported by the POV-Ray Team.  There are strict rules under
*  which you are permitted to use this file.  The rules are in the file
*  named POVLEGAL.DOC which should be distributed with this file.
*  If POVLEGAL.DOC is not available or for more info please contact the POV-Ray
*  Team Coordinator by email to team-coord@povray.org or visit us on the web at
*  http://www.povray.org. The latest version of POV-Ray may be found at this site.
*
* This program is based on the popular DKB raytracer version 2.12.
* DKBTrace was originally written by David K. Buck.
* DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
*
*****************************************************************************/

#include "frame.h"
#include "povray.h"
#include "vector.h"
#include "povproto.h"
#include "bbox.h"
#include "hfield.h"
#include "matrices.h"
#include "objects.h"
#include "planes.h"
#include "quadrics.h"
#include "texture.h"

#ifdef MotionBlurPatch
#include "mblur2.h"

extern int TimeStamp;


/*****************************************************************************
* Local preprocessor defines
******************************************************************************/




/*****************************************************************************
* Static functions
******************************************************************************/

static int All_Motion_Blur2_Intersections (OBJECT *Object, RAY *Ray, ISTACK *Depth_Stack);
static int Inside_Motion_Blur2 (VECTOR point, OBJECT *Object);
static MOTION_BLUR2 *Copy_Motion_Blur2 (OBJECT *Object);
static void Translate_Motion_Blur2 (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans);
static void Rotate_Motion_Blur2 (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans);
static void Scale_Motion_Blur2 (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans);
static void Transform_Motion_Blur2 (OBJECT *Object, TRANSFORM *Trans);
static void Destroy_Motion_Blur2 (OBJECT *Object);
static void Invert_Motion_Blur2 (OBJECT *Object);


/*****************************************************************************
* Local variables
******************************************************************************/

METHODS Motion_Blur2_Methods =
{
  All_Motion_Blur2_Intersections,
  Inside_Motion_Blur2, NULL /*Normal*/, Default_UVCoord /* UVCoord */,
  (COPY_METHOD)Copy_Motion_Blur2,
  Translate_Motion_Blur2, Rotate_Motion_Blur2,
  Scale_Motion_Blur2, Transform_Motion_Blur2, Invert_Motion_Blur2, Destroy_Motion_Blur2
};


/*****************************************************************************
*
* FUNCTION
*
*   All_Motion_Blur2_Intersections
*
* INPUT
*   
* OUTPUT
*   
* RETURNS
*   
* AUTHOR
*
*   
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   Sep 1994 : Added code to count intersection tests. [DB]
*
******************************************************************************/

static int All_Motion_Blur2_Intersections (OBJECT *Object, RAY *Ray, ISTACK *Depth_Stack)
{
  int Found;
  DBL delta;
  OBJECT *Current_Sib, *Clip;
  ISTACK *Local_Stack;
  INTERSECTION *Sibling_Intersection;
  TRANSFORM Trans;
  TRANSFORM InvTrans;
  MATRIX Mtx;
  MOTION_BLUR2 *o;
  int i,j;
  RAY New_Ray;
  RAY New_Ray2;
  DBL len;

  o = (MOTION_BLUR2*)Object;

  if (!Ray_In_Bound (Ray, o->Bound))
  {
    return FALSE;
  }

  Increase_Counter(stats[Ray_Motion_Blur_Tests]);

  Current_Sib = ((MOTION_BLUR2 *)Object)->Children;

  if(opts.motionBlurCount>1)
  {
    /* motion blur enabled - let's do this */
    if (!TimeStamp) 
    {
      Local_Stack = open_istack();
      Found = FALSE;
      for (TimeStamp = 1; TimeStamp <= opts.motionBlurCount; TimeStamp++)
      {
        if (All_Intersections (Object, Ray, Local_Stack))
        {
          Found = TRUE;
          break;
        }
      }

      if (!Found) 
      {
        delta = 0.5;
        TimeStamp = 0;
      }
      else
      {
        Current_Sib->TimeStamp = TimeStamp;
        delta = (DBL)(TimeStamp-1)/(opts.motionBlurCount-1);
        if (delta>1.0) delta = 1.0;
        if (delta<0.0) delta = 0.0;
      }

      /*TimeStamp = 0;*/

      Found = FALSE;
      close_istack (Local_Stack);
    }
    else
    {
      /* motion blur type 1 */
      Current_Sib->TimeStamp = TimeStamp;
      delta = (DBL)(TimeStamp-1)/(opts.motionBlurCount-1);
      if (delta>1.0) delta = 1.0;
      if (delta<0.0) delta = 0.0;
    }
  }
  else
  {
    delta = 0.5;
    Current_Sib->TimeStamp = 0;
  }


  if(Current_Sib->TimeStamp != o->LastTimeStamp)
  {
    /* linearly interpolate the matrix */
    for(i=0; i<4; i++)
      for(j=0; j<4; j++)
        Trans.matrix[i][j] = Mtx[i][j] = o->Trans1.matrix[i][j]*delta+o->Trans2.matrix[i][j]*(1.0-delta);

    Compute_Matrix_Transform(&Trans,Mtx);

    /* this has a problem with the POST_DONE flag */

    /* undo the last transform */
    Transform_Object (Current_Sib, &o->LastTrans);
    /* now transform it to this one */
    Transform_Object(Current_Sib,&Trans);

    for(i=0; i<4; i++)
      for(j=0; j<4; j++)
      {
        o->LastTrans.matrix[i][j] = Trans.inverse[i][j];
        o->LastTrans.inverse[i][j] = Trans.matrix[i][j];
      }

    o->LastTimeStamp = Current_Sib->TimeStamp;
  }

  if (Ray_In_Bound (Ray, Current_Sib->Bound))
  {
    Found = All_Intersections (Current_Sib, Ray, Depth_Stack);
  }

  if (Found)
  {
    Increase_Counter(stats[Ray_Motion_Blur_Tests_Succeeded]);
  }

  return (Found);
}



/*****************************************************************************
*
* FUNCTION
*
*   Inside_Motion_Blur2
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

static int Inside_Motion_Blur2 (VECTOR IPoint, OBJECT *Object)
{
  OBJECT *Current_Sib;

  Current_Sib = ((MOTION_BLUR2 *)Object)->Children;

  if (Inside_Object (IPoint, Current_Sib))
  {
    return (TRUE);
  }

  return (FALSE);
}



/*****************************************************************************
*
* FUNCTION
*
*   Translate_Motion_Blur2
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Translate_Motion_Blur2 (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans)
{
  Translate_Object(((MOTION_BLUR2 *) Object)->Children, Vector, Trans);

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Rotate_Motion_Blur2
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Rotate_Motion_Blur2 (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans)
{
  Rotate_Object(((MOTION_BLUR2 *) Object)->Children, Vector, Trans);

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Scale_Motion_Blur2
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Scale_Motion_Blur2 (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans)
{
  Scale_Object(((MOTION_BLUR2 *) Object)->Children, Vector, Trans);

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Transform_Motion_Blur2
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Transform_Motion_Blur2 (OBJECT *Object, TRANSFORM *Trans)
{
  Transform_Object (((MOTION_BLUR2 *) Object)->Children, Trans);

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Invert_Motion_Blur2
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Invert_Motion_Blur2 (OBJECT *Object)
{
  Invert_Object (((MOTION_BLUR2 *)Object)->Children);

  Invert_Flag(Object, INVERTED_FLAG);
}





/*****************************************************************************
*
* FUNCTION
*
*   Create_Motion_Blur2
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

MOTION_BLUR2 *Create_Motion_Blur2 ()
{
  MOTION_BLUR2 *New;

  New = (MOTION_BLUR2 *)POV_MALLOC(sizeof (MOTION_BLUR2), "motion blur");

  INIT_OBJECT_FIELDS(New, MOTION_BLUR_OBJECT+COMPOUND_OBJECT, &Motion_Blur2_Methods)

  New->Children = NULL;

  MIdentity(New->LastTrans.matrix);
  MIdentity(New->LastTrans.inverse);
  New->LastTimeStamp = 0;

  return (New);
}



/*****************************************************************************
*
* FUNCTION
*
*   Copy_Motion_Blur2
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

static MOTION_BLUR2 *Copy_Motion_Blur2 (OBJECT *Object)
{
  MOTION_BLUR2 *New;

  New = (MOTION_BLUR2 *)POV_MALLOC(sizeof (MOTION_BLUR2), "motion blur");

  *New = *(MOTION_BLUR2 *)Object;

  New->Children = Copy_Object (((MOTION_BLUR2 *)Object)->Children);

  return (New);
}



/*****************************************************************************
*
* FUNCTION
*
*   Destroy_Motion_Blur2
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Destroy_Motion_Blur2 (OBJECT *Object)
{
  Destroy_Object (((MOTION_BLUR2 *) Object)->Children);

  POV_FREE (Object);
}



/*****************************************************************************
*
* FUNCTION
*
*   Compute_Motion_Blur2_BBox
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   POV-Ray Team
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*
******************************************************************************/

void Compute_Motion_Blur2_BBox (OBJECT *Object)
{
  DBL Old_Volume, New_Volume;
  VECTOR NewMin, NewMax, TmpMin, TmpMax;
  OBJECT *Sib;
  MOTION_BLUR2 *o;
  TRANSFORM InvTrans1, InvTrans2;
  int i,j;

  o = (MOTION_BLUR2*)Object;

  for (i=0; i<4; i++)
    for (j=0; j<4; j++)
    {
      InvTrans1.inverse[i][j] = o->Trans1.matrix[i][j];
      InvTrans1.matrix[i][j] = o->Trans1.inverse[i][j];
      InvTrans2.inverse[i][j] = o->Trans2.matrix[i][j];
      InvTrans2.matrix[i][j] = o->Trans2.inverse[i][j];
    }

  Make_Vector(NewMin,  BOUND_HUGE,  BOUND_HUGE,  BOUND_HUGE);
  Make_Vector(NewMax, -BOUND_HUGE, -BOUND_HUGE, -BOUND_HUGE);

  Sib = ((MOTION_BLUR2 *)Object)->Children;

  for(i=0; i<=1; i++)
  {
    if(i)
      Recompute_BBox(&Sib->BBox, &o->Trans2);
    else
      Recompute_BBox(&Sib->BBox, &o->Trans1);

    Make_min_max_from_BBox(TmpMin, TmpMax, Sib->BBox);

    NewMin[X] = min(NewMin[X], TmpMin[X]);
    NewMin[Y] = min(NewMin[Y], TmpMin[Y]);
    NewMin[Z] = min(NewMin[Z], TmpMin[Z]);
    NewMax[X] = max(NewMax[X], TmpMax[X]);
    NewMax[Y] = max(NewMax[Y], TmpMax[Y]);
    NewMax[Z] = max(NewMax[Z], TmpMax[Z]);

    if(i)
      Recompute_BBox(&Sib->BBox, &InvTrans2);
    else
      Recompute_BBox(&Sib->BBox, &InvTrans1);
  }
 
  if ((NewMin[X] > NewMax[X]) || (NewMin[Y] > NewMax[Y]) || (NewMin[Z] > NewMax[Z]))
  {
    Warning(0, "Degenerate motion blur bounding box (not used!).\n");
  }
  else
  {
    New_Volume = (NewMax[X] - NewMin[X]) * (NewMax[Y] - NewMin[Y]) * (NewMax[Z] - NewMin[Z]);

    BOUNDS_VOLUME(Old_Volume, Object->BBox);

    if (New_Volume < Old_Volume)
    {
      Make_BBox_from_min_max(Object->BBox, NewMin, NewMax);

      /* Beware of bounding boxes to large. */

      if ((Object->BBox.Lengths[X] > CRITICAL_LENGTH) ||
          (Object->BBox.Lengths[Y] > CRITICAL_LENGTH) ||
          (Object->BBox.Lengths[Z] > CRITICAL_LENGTH))
      {
        Make_BBox(Object->BBox, -BOUND_HUGE/2, -BOUND_HUGE/2, -BOUND_HUGE/2,
          BOUND_HUGE, BOUND_HUGE, BOUND_HUGE);
      }
    }
  }
}

#else
 static  char dummy[2];
#endif


